
/** @file vdif_speana test Spectrum check of VDIF data 
 * @author M.Sekido
 * @Version $Rev: 56 $
 * @Date  $Date: 2013-07-03 20:40:14 +0900 (Wed, 03 Jul 2013) $
 * $Id: vdif_speana.cpp 56 2013-07-03 11:40:14Z vlbi $ 
 */

#include <iostream>
#include <string>
#include <vector>
using namespace std;

#include <cstdio>
#include <cstdlib>

#include "VDIF_Header.hpp"
#include "VDIF_Frame.hpp"
#include "DAS_info.hpp"
#include "VsiData.hpp"
#include "pltutl.hpp"
#include "fftw_utl.hpp"
#include "utlty.hpp"
#include "conf_utl.hpp"


extern "C"{
#include <unistd.h>
}

/** General Configulation data */
Config_set vdif_conf;

const int MAXCH=16;

template <typename T> class dGen{
  /** value to keeping state */
  T r;
  /** delta increment value  */
  T dt;
public:
  /** Default constructor */
  dGen():r(0),dt(1){}
  /** Constructor with initialization */
  dGen(T r0, T d):r(r0),dt(d){}
  /** Setting the initial value */
  void Set_init(const T r0, const T del=0){r = r0; dt=del;}
  /** Setting the delta value */
  void Set_delt(const T d){dt=d;}
  /** Opearator for returning generated value **/ 
  T operator()(){   r += dt; return (r-dt);  }
};

/**! "main_cmn.hpp": Common Declraration of static parameters. 
 * Should be included in main file only. */
#include "main_cmn.hpp"


void usage(){
  cerr << "Usage: " << me << " -f <filename> [-c <nch>] [-t <sec>] [-l]\n";
  cerr << "      [-n <nfft>] [-b <bit encode>] [-g <PGPLOT out>]\n";
  cerr << "-c <ch>  : Select <ch> data.(default=-1 means all channels.)\n";
  cerr << "-t <sec> : float Second for integration. (default=0.1 sec.)\n";
  cerr << "-l       : Linear scalse. (default=dB scale 10*log10) \n";
  cerr << "-n <nfft>: Number of FFT points. (default=2048) \n";
  cerr << "-g       : Pgplot output to file (default=/xw) \n";
  cerr << "-b       : one of [ADS2000/ADS3000/SignMag/RevSign]. (default=ADS2000) \n";
  cerr << "           ADS2000: 00,01,10,11 from bottom to top level.\n";
  cerr << "           ADS3000: 00,10,01,11 from bottom to top level.\n";
  cerr << "           SignMag: 01,00,10,11 from bottom to top level.\n";
  cerr << "           RevSgn : 10,00,01,11 from bottom to top level.\n";
  cerr << " Compute Spectrum for monitor.\n";
  exit(1);
}

/** Averaging data.
 * \return navg Number of averaging count;
 */
const int Avg_spectrum(string &fname, const float tdur, const int nfft, 
		       const VsiData::BitEnc_t code_type,
		       //const int ch, const VsiData::BitEnc_t code_type,
		       vector <float> avg_arr[], vector <VDIF_Hdr>& hdrarr ){

  VDIF_Frame  vdif;
  if( ! vdif.ROpen(fname)){
    cerr << "Error in opening file: " << fname << endl;
    exit(1);
  }


  /**! Reading Header info  */  
  //  long nsec = 
  size_t nsec = 10; // sample 10 sec.
  // First reading data to confirm Srate 
  cerr << "Reading every sec...";
  int nframe = vdif.Read_every_sec_hdr(hdrarr, nsec, VDIF_Frame::QUICK);
  if(nframe < 1){
    cerr << "E Failed to find VDIF Header.\n";
    exit(1);
  }
  cerr << hdrarr.size() << " sec. read." << endl;
  DAS_Info das = vdif.DAS_Inf();// Sampling info
  int nch = das.Nch();
  int qbit = das.Qbit();

//   if(ch >= nch){
//     fprintf(stderr, "E Avg_spectrum(): Requested ch(%d) >= Nch(%d)\n", ch, nch);
//     return -1;
//   }
  if(nch > MAXCH){
    fprintf(stderr, "E Avg_spectrum(): Nch(%d) in the data > MAXCH(%d)\n", nch, MAXCH);
    return -1;
  }

  size_t datasize = hdrarr[0].Framesize() - hdrarr[0].Headsize();
  long pnt_frame = datasize *8 /qbit/nch; // #points/frame
  float sec_frame = (float)datasize/4/das.Speed(); // seconds per frame
  //  float dt_fft = (float)nfft/(float)das.Srate();  // seconds per fft.
  int tframe = (int)(tdur / sec_frame);   // Total #frames requested. 

  VsiData vsidat(&das);
  vsidat.Make_Table(code_type); // Make Decode Table 
  
  vector <uint8_t> datbuf;
  vector <VDIF_Hdr> cur_hdrarr;
  vector <FComplex> cmplx_out(nfft/2); // Output FFT Complex 

  vector <float>  amp_arr(nfft/2), phs_arr(nfft/2); 
  
  // Number of for reading
  //  nframe = (tdur >1.0 ? (int)(1./sec_frame) : tframe);
  nframe = (tframe > 10 ? 10 : tframe);
  //  long npt_read = nframe * pnt_frame;
  //  vector <Fftwf> fftwdat;
  Fftwf fftwdat[MAXCH];

  vdif.Rewind(); // Rewind the File ponter   
  long rframe = vdif.Read_frame(cur_hdrarr, datbuf, nframe);
  if(rframe < nframe){ /*Check reading*/
    cerr << "W Reading data(" << rframe << "frames)< nframe(" 
	 << nframe  << ")" << endl;
  }
  long n_available = vsidat.Interpret(datbuf,  pnt_frame);
  if(n_available < 1){ 
    fprintf(stderr,"E Failed VsiDat::Interpret(): ret(%ld) < npt(%ld)\n", 
	    n_available,  pnt_frame);
    return -1;
  }

  vector <const float*> pf;
  vector <fftwf_complex*> pfftw_in;
  int npch=nch; // Number of processing channels
  
  //  if(ch == -1){  npch = nch;  }

  pf.resize(npch); 
  //  fftwdat.resize(npch); 
  pfftw_in.resize(npch); // all channels 
  for(int i=0; i<npch; i++){
    pf[i] = vsidat.pfValArr(i); // Pointer to data
    fftwdat[i].Init(nfft);
    pfftw_in[i]  = fftwdat[i].In();
    if(avg_arr[i].size() < (size_t)nfft/2) avg_arr[i].resize(nfft/2);
  }
  
  int navg =0;
  long frm_count = rframe;
  cerr << "Reading Frame:" << frm_count  << "\r";

  
  while( frm_count < tframe ){
    for(int i=0; i<npch; i++){
      pfftw_in[i]  = fftwdat[i].In();
      memset((void*)(*pfftw_in[i]),0,sizeof(fftwf_complex)*nfft);
    }
    /** FILL FFT input array **/
    if(nfft < n_available ){
      for(int i=0; i<npch; i++){
	fftwdat[i].Set_in(pf[i]);
	pf[i] += nfft;
      }
      n_available -= nfft;
    }else{
      long n_need = nfft - n_available;
      for(int i=0; i<n_available; i++ ){
	for(int ich=0; ich<npch; ich++){
	  (*pfftw_in[ich])[0] = *pf[ich];
	  pf[ich]++;  pfftw_in[ich]++;
	}
      }
      n_available =0;

      while(n_need > 0){
	if(  (rframe = vdif.Read_frame(cur_hdrarr, datbuf, nframe)) < 1){
	  cout << endl;
	  break;
	}
	frm_count += rframe;  // cerr << frm_count << " ";
	cerr << "Reading Frame:" << frm_count  << "\r";
	if(rframe < nframe){ /*Check reading*/
	  cerr << "W Reading data(" << rframe << "frames)< nframe(" 
	       << nframe  << ")" << endl;
	}
	n_available = vsidat.Interpret(datbuf,  pnt_frame);
// 	if(n_available < 1){ 
// 	  fprintf(stderr,"E Failed VsiDat::Interpret(): ret(%ld) < npt(%ld)\n",
// 		  n_available, rframe * pnt_frame);
// 	  break;
// 	}
	for(int ich=0; ich<npch; ich++) pf[ich] = vsidat.pfValArr(ich);

	if(n_need > n_available){
	  for(int i=0; i<n_available; i++)
	    for(int ich=0; ich<npch; ich++){
	      (*pfftw_in[ich])[0] = *pf[ich];
	      pf[ich]++; pfftw_in[ich]++;
	    }
	  n_need -= n_available; n_available=0;
	}else{
	  for(int i=0; i<n_need; i++)
	    for(int ich=0; ich<npch; ich++){
	      (*pfftw_in[ich])[0] = *pf[ich];
	      pf[ich] ++; pfftw_in[ich] ++;
	    }
	  n_available -= n_need; n_need =0;
	} // else 
      } // End while
    } // else
    /** FILL FFT input array END **/
    
    navg++;
    float fact1 = 1./(float)navg;
    float fact2 = (float)(navg-1)/(float)navg;
    for(int ich=0; ich<npch; ich++){
      fftwdat[ich].Exec(-1);
      //  memcpy(&cmplx_out[0], fftwdat[ich].Out() + nfft/2, sizeof(fftwf_complex)*nfft/2);
      memcpy(&cmplx_out[0], fftwdat[ich].Out(), sizeof(fftwf_complex)*nfft/2);
      FCarr2AParr(cmplx_out, amp_arr, phs_arr);
      vector <float>::iterator i1 = avg_arr[ich].begin();
      vector <float>::iterator i2 = amp_arr.begin();
      for(; i1 < avg_arr[ich].end(); i1++, i2++)  *i1 = *i1*fact2 + *i2*fact1;
    } //  for ich
  } // while tdur 
  
  return navg;

}


int main(int argc, char *argv[]){
  me = (me = strrchr(argv[0], '/')) == NULL ? argv[0] : me+1; /* My name */
  enum Plot_t {Linear, dBscale};
  
  string fname;
  int nfft=2048;
  float tdur=0.1;  // (sec)
  VsiData::BitEnc_t  code_type = VsiData::ADS2000_2bit;
  int ch = -1; // Channel number 
  Plot_t plot_type = dBscale ; // 10*log10 dB scale
  string pgdev("/xw");

  int k=0;
  while (k >= 0) { 
    switch (k = getopt(argc, argv, "hf:c:t:n:b:g:l")) { 
    case 'h':  usage(); break;
    case 'c':    ch = atoi(optarg);      break;
    case 'f':    fname = optarg;      break;
    case 'n':    nfft = atoi(optarg);      break;
    case 'l':    plot_type = Linear;  break; // dB scale
    case 't':    tdur = atof(optarg);      break;
    case 'g':    pgdev = optarg;      break;
    case 'b': 
      if( strncmp(optarg, "ADS3000",7)==0) code_type = VsiData::ADS3000_2bit;
      if( strncmp(optarg, "SignMag",7)==0) code_type = VsiData::SignMag_2bit;
      if( strncmp(optarg, "RevSign",7)==0) code_type = VsiData::RevSgnMag_2bit;
      break;
      //    case 'd': vdif_hdr.Set_dbg(atoi(optarg)); break;
    case '?' : /* Error:  character not in list */ 
      fprintf(stderr, /* Yes */ 
	      "%s WARNING:  Unknown option on command line \n", me); 
      usage();
      break; 
    case ':' : /* Error:  missing parameter */ 
      fprintf(stderr, "%s WARNING:  Missing number on command line \n", me); 
      usage();
      break; 
    }
  }

  if(fname.empty()) usage();

  vector <float>    avg_arr[MAXCH];
  vector <VDIF_Hdr> hdrarr;

  fprintf(stdout,"Nfft=%4d pnts, Integration= %.2f sec. \n", nfft, tdur);
  // Get Averaged Spectrum
  if(Avg_spectrum(fname, tdur, nfft, code_type, avg_arr, hdrarr ) < 0){
    cerr << "E Failed in Avg_spectrum() \n";
    exit(1);
  }
  fprintf(stdout,"Data Info: %s\n", hdrarr[0].Str_hd(20).c_str());

  float df = (float)hdrarr[0].Srate()/nfft * 1.e-6; // (MHz)
  vector <float> xarr(nfft/2);
  dGen<float>g1(0,df);
  generate( xarr.begin(), xarr.end(), g1);

  int nch  = hdrarr[0].Nch();
  int qbit = hdrarr[0].Qbit();
  // Plot data 
  Plt_inf pltinf;
  char str1[2][256],str2[256], str3[256];
  sprintf(str1[0],"\\fr%s (VDIF Ver.%s)",fname.c_str(),
	  hdrarr[0].Version().c_str());
  //  if(hdrarr[0].ext_dat_ver == 1) // Extension NICT version
  sprintf(str1[1],"\\fr%dMsps/%dbit/%dch",
	  hdrarr[0].Srate()/1000000,qbit,nch);
  sprintf(str2,"\\frIntg:%.2fsec, df=%.2fkHz(n=%d)",tdur, df*.1e3, nfft);
  cpgopen( pgdev.c_str() ); 
  switch(plot_type){
  case Linear:   pltinf.Set_label("\\frFreq. (MHz)", "\\frPower"); break;
  case dBscale:
  default:
    pltinf.Set_label("\\frFreq. (MHz)", "\\frPower [dB]"); 
    pltinf.Set_ymax(30);
    pltinf.Set_ymin(-10);
    pltinf.Set_ytick(10);
    break;
  }

  
  int npch =1;
  if(ch == -1) npch = nch;
  switch(npch){
  case 2:
    cpgsubp(2,1);
    break;
  case 4:
    cpgsubp(2,2);
    break;
  case 8:
    cpgsubp(4,2);
    break;
  case 16:
    cpgsubp(4,4);
    break;
  case 1:
  default:
    break;
  }
  cpgask(1);
  vector <float> dbarr;

  for(int ich=0; ich<npch; ich++) {
    cpgpage();
    cpgsvp(0.15, 0.9, 0.15, 0.9);  
    if(ch != -1){
      sprintf(str3,"\\frCH=%d", ch);
      if(plot_type == Linear)  
	pltinf.Plot1(xarr,avg_arr[ch]);
      else {
	dbarr = val2dB(avg_arr[ch]);
	pltinf.Plot1(xarr,dbarr);
      }
    }else{
      if(plot_type == Linear)
	pltinf.Plot1(xarr,avg_arr[ich]);
      else{         
	dbarr = val2dB(avg_arr[ich]);
	pltinf.Plot1(xarr,dbarr);
      }
      sprintf(str3,"\\frCH=%d", ich);
    }
    cpgsch(1.5);
    cpgmtxt("T", -1, 0.9,1,str3);
  }

  cpgsubp(1,1);
  cpgsch(0.8);
  cpgsvp(0.1,0.9,0.1,0.9);
  //  if(ich==0){
    cpgmtxt("T", 4, 0,0, str1[0]);
    cpgmtxt("T", 4, 0.9,1, str1[1]);
    cpgmtxt("T", 3, 0,0, str2);
    //  }

  //  cpgpage();
  cpgclos();

}
